from pathlib import Path
from typing import Dict, Callable, Union, Optional, TYPE_CHECKING

from yaml import safe_load
if TYPE_CHECKING:
    from aiohttp import ClientSession as AioSession, ClientResponse as AioResponse
    from requests import Session as ReqSession, Response as ReqResponse


PathT = Union[Path, str]
CallableDefaultT = Callable[[], Optional[str]]
DefaultValueT = Union[str, CallableDefaultT]
OptionsT = Dict[str, str]


class FormField:
    def __init__(self, name: str, default: DefaultValueT = None, options: OptionsT = None, required: bool = False):
        self.name: str = name
        self._default: Optional[DefaultValueT] = default
        self.options: Optional[OptionsT] = options
        self.required: bool = required

    def __repr__(self) -> str:
        s = '<InputField' if self.options is None else '<SelectField'
        s += f': name="{self.name}"'
        if self.required:
            s += ', required=True'
        if self._default is not None:
            s += f', default="{self.default}"'
        return s + '>'

    def __str__(self) -> str:
        return repr(self)

    @property
    def default(self) -> Optional[str]:
        return self._default() if callable(self._default) else self._default

    def valid_option(self, option: str) -> bool:
        return self.options is None or option in self.options.keys() or option in self.options.values()

    def clean_value(self, value: str) -> str:
        if self.options is None or value in self.options.keys():
            return value
        # Try to find an unambiguous match in the visible option texts:
        key = None
        for option_key, option_text in self.options.items():
            if option_text == value:
                if key is not None:
                    raise LookupError(f'More than one options key has the text "{value}" in {self}')
                key = option_key
        if key is not None:
            return key
        raise ValueError(f'"{value}" is not a valid option for {self}')


class Form:
    def __init__(self, definition: Dict[str, Dict], full_payload: bool = True, url: str = None):
        """
        Creates a form instance from a definition dictionary. Each element in the dictionary must define a field.

        Args:
            definition:
                The field definition dictionary
            full_payload (optional):
                Defines the behavior during payload validation/creation.
                If True (default), the fields' predefined default values will be inserted into the payload,
                wherever no value (or `None`) is explicitly passed for that field.
                If False, fields for which no values were explicitly passed, will not be included in the payload.
                The exception is required fields; these will cause an error, if no value is provided.
            url (optional):
                Can be set in advance to the url that requests using this form's payload should be made to.
        """
        self._fields: Dict[str, FormField] = {alias: FormField(**field_def) for alias, field_def in definition.items()}
        self.full_payload_always: bool = full_payload
        self.url: Optional[str] = url

    @property
    def fields(self) -> Dict[str, FormField]:
        return self._fields.copy()

    @property
    def required_fields(self) -> Dict[str, FormField]:
        return {alias: field for alias, field in self._fields.items() if field.required}

    def get_payload(self, **kwargs: str) -> Dict[str, str]:
        """
        Creates a request payload from the form's fields as a dictionary,
        where the keys represent the name attributes of the form fields and the values the actual values to be passed.

        This payload can then be used for example in a GET request in the form of url parameters (after encoding it
        in the appropriate string form) or in a regular POST request.

        Args:
            kwargs (optional):
                Every key must correspond to a key in the internal dictionary of fields;
                otherwise that key-value-pair is ignored.
                Values will be passed into the payload (if they pass validation).
                Select fields with predefined options will only allow one of the options to be passed.
                If `None` is passed as a value, the corresponding field's default value will be used in the payload.

        Returns:
            Validated name-value-mapping to be used for HTTP requests from the form's fields.
        """
        payload = {}
        for key, field in self.fields.items():
            if key in kwargs.keys():
                value = kwargs[key]
                if value is None:
                    payload[field.name] = field.default
                else:
                    payload[field.name] = field.clean_value(value)
            elif field.required:
                raise ValueError(f"`{key}` is a required field, but no argument was passed.")
            elif self.full_payload_always:
                payload[field.name] = field.default
        return payload

    async def post_aio(self, _aiohttp_session_obj: 'AioSession' = None, **kwargs: str) -> 'AioResponse':
        """
        Uses `aiohttp` to perform a POST request to `.url` with the form's payload generated using `kwargs`.

        Args:
            _aiohttp_session_obj (optional):
                Can be set to a pre-existing `aiohttp.ClientSession` instance that should be used for the request.
            kwargs:
                Passed directly into `.get_payload`.

        Returns:
            The `aiohttp.ClientResponse` object from the request.
        """
        if self.url is None:
            raise AttributeError("`url` attribute not set")
        from aiohttp import ClientSession, ClientResponse
        from webutils import in_async_session

        @in_async_session
        async def post(url: str, data: dict, session: ClientSession = None) -> ClientResponse:
            async with session.post(url, data=data) as response:
                return response
        return await post(self.url, self.get_payload(**kwargs), session=_aiohttp_session_obj)

    def post_req(self, _requests_session_obj: 'ReqSession' = None, **kwargs: str) -> 'ReqResponse':
        """
        Uses `requests` to perform a POST request to `.url` with the form's payload generated using `kwargs`.

        Args:
            _requests_session_obj (optional):
                Can be set to a pre-existing `requests.Session` instance that should be used for the request.
            kwargs:
                Passed directly into `.get_payload`.

        Returns:
            The `requests.Response` object from the request.
        """
        if self.url is None:
            raise AttributeError("`url` attribute not set")
        if _requests_session_obj is not None:
            return _requests_session_obj.post(self.url, data=self.get_payload(**kwargs))
        from requests import post
        return post(self.url, data=self.get_payload(**kwargs))


def load_form(*def_paths: PathT, full_payload: bool = True) -> Form:
    """
    Creates a form instance from an arbitrary number of definition files in YAML format.

    Args:
        def_paths:
            Each element must be a path to a readable YAML file containing a valid form definition.
        full_payload (optional):
            Passed directly into the Form constructor.
    """
    definition = {}
    for path in def_paths:
        with open(path, 'r') as f:
            definition.update(safe_load(f))
    return Form(definition, full_payload=full_payload)
